#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <pwd.h>
#include <grp.h>
#include <libgen.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/sysmacros.h>

#include "lib.h"
#include "fs.h"
#include "archive.h"

int checksum(char *block, int sgn) {
	size_t i;
	int result;
	result = 0;
	for (i = 0; i < 512; i++) {
		if (i >= 148 && i < 156)
			result += 32;
		else if (sgn)
			result = (result + (char)(block[i])) & 0x3ffff;
		else
			result = (result + (unsigned char)(block[i])) & 0x3ffff;
	}
	return result;
}

void output_infoblock(FILE *out, char *outfn, char *path, mode_t mode,
		uid_t uid, gid_t gid, off_t size, time_t time, char block_type,
		char *target, char *username, char *groupname, dev_t dev) {
	unsigned char block[512];
	size_t i;
	for (i = 0; i < 512; i++)
		block[i] = 0;
	strncpy(block, path, 100);
	snprintf(block + 100, 8, "%07o", mode);
	snprintf(block + 108, 8, "%07o", uid);
	snprintf(block + 116, 8, "%07o", gid);
	snprintf(block + 124, 12, "%011o", size);
	snprintf(block + 136, 12, "%011o", time);
	block[156] = block_type;
	if (target)
		strncpy(block + 157, target, 100);
	sprintf(block + 257, "ustar");
	sprintf(block + 263, "00");
	if (username)
		strncpy(block + 265, username, 32);
	if (groupname)
		strncpy(block + 297, groupname, 32);
	snprintf(block + 329, 8, "%06o ", major(dev));
	snprintf(block + 337, 8, "%06o ", minor(dev));
	snprintf(block + 148, 8, "%06o ", checksum(block, 0));
	xwrite(out, outfn, block, 512);
}

void output_long_link(FILE *out, char *filename, char *path) {
	size_t i, len;
	len = strlen(path);
	output_infoblock(out, filename, "././@LongLink", 0, 0, 0, strlen(path), 0,
		'L', NULL, "root", "root", 0);
	xwrite(out, filename, path, len);
	fill_bytes(out, filename, (512 - len % 512) % 512);
}

int path_is_untracked(char *path, struct stringlist *untracked) {
	struct stringlist *curr;
	char *s, *t;
	for (curr = untracked; curr; curr = curr->next) {
		s = strdup(path);
		if (!s) {
			fprintf(stderr, "%s: %s: out of memory\n", progname, path);
			exit(1);
		}
		while (1) {
			if (paths_equal(s, curr->s)) {
				free(s);
				return 1;
			}
			if (!strcmp(s, ".") || !strcmp(s, "/")) {
				free(s);
				break;
			}
			t = strdup(dirname(s));
			if (!t) {
				fprintf(stderr, "%s: %s: out of memory\n", progname, path);
				exit(1);
			}
			free(s);
			s = t;
		}
	}
	return 0;
}

int prepare_directory(struct path_info **pi, char *name, int linkfirst) {
	DIR *dir;
	struct dirent *de;
	int result;
	result = 0;
	dir = opendir(name);
	if (!dir) {
		error(name);
		return -1;
	}
	for (de = readdir(dir); de; de = readdir(dir))
		if (strcmp(de->d_name, ".") && strcmp(de->d_name, "..") &&
				prepare_path(pi, name, de->d_name, linkfirst))
			result = -1;
	closedir(dir);
	return result;
}

int prepare_single_path_info(struct path_info **pi, char *name,
		int linkfirst) {
	struct path_info **cpi;
	struct stat st;
	int result, link;
	link = 0;
	if (!path_is_safe(name)) {
		fprintf(stderr, "%s: %s: invalid path\n", progname, name);
		return -1;
	}
	if (lstat(name, &st)) {
		error(name);
		return -1;
	}
	for (cpi = pi; *cpi; cpi = &((*cpi)->next)) {
		if ((*cpi)->inode == st.st_ino && (*cpi)->device == st.st_dev) {
			if (linkfirst)
				(*cpi)->link = 1;
			else
				link = 1;
		}
	}
	*cpi = malloc(sizeof(struct path_info));
	if (!*cpi) {
		fprintf(stderr, "%s: %s: out of memory\n", progname, name);
		return -1;
	}
	(*cpi)->next = NULL;
	(*cpi)->name = name;
	(*cpi)->device = st.st_dev;
	(*cpi)->inode = st.st_ino;
	(*cpi)->link = link;
	return !!S_ISDIR(st.st_mode);
}

int prepare_path(struct path_info **pi, char *parent, char *path,
		int linkfirst) {
	char *name;
	if (parent)
		name = construct_child_path(parent, path);
	else
		name = strdup(path);
	if (!name) {
		fprintf(stderr, "%s: %s/%s: out of memory\n", progname, parent, path);
		return -1;
	}
	switch (prepare_single_path_info(pi, name, linkfirst)) {
	case -1:
		free(name);
		return -1;
	case 0:
		return 0;
	case 1:
		return prepare_directory(pi, name, linkfirst);
	}
}

int output_path(FILE *out, char *outfn, struct path_info *pi,
		struct path_info *opi, struct stringlist *untracked, uid_t uid,
		char *uname, gid_t gid, char *gname) {
	unsigned char pathbuf[101];
	unsigned char tgtbuf[101];
	unsigned char username[33];
	unsigned char groupname[33];
	char *target, *userptr, *groupptr;
	char blocktype;
	struct stat st;
	struct passwd *pw;
	struct group *gr;
	FILE *in;
	struct path_info *cpi;
	off_t size;
	dev_t dev;
	size_t pathlen;
	ssize_t len;
	if (lstat(pi->name, &st)) {
		error(pi->name);
		return -1;
	}
	pathlen = strlen(pi->name);
	if (S_ISDIR(st.st_mode))
		pathlen++;
	if (pathlen > 100)
		output_long_link(out, outfn, pi->name);
	target = NULL;
	size = 0;
	if (S_ISDIR(st.st_mode))
		snprintf(pathbuf, 101, "%s/", pi->name);
	else
		snprintf(pathbuf, 101, "%s", pi->name);
	blocktype = '0';
	if (pi->link) {
		for (cpi = opi; cpi && cpi != pi; cpi = cpi->next) {
			if (cpi->inode == st.st_ino && !cpi->link) {
				blocktype = '1';
				snprintf(tgtbuf, 101, cpi->name);
				target = tgtbuf;
				break;
			}
		}
	}
	if (blocktype != '1') {
		if (S_ISDIR(st.st_mode)) {
			blocktype = '5';
		} else if (S_ISREG(st.st_mode)) {
			size = st.st_size;
			blocktype = '0';
		} else if (S_ISLNK(st.st_mode)) {
			blocktype = '2';
			if (st.st_size > 100) {
				fprintf(stderr, "%s: %s: link target too long\n", progname,
					pi->name);
				return -1;
			}
			len = readlink(pi->name, tgtbuf, 100);
			if (len < 0) {
				error(pi->name);
				return -1;
			}
			tgtbuf[len] = 0;
			target = tgtbuf;
		} else if (S_ISCHR(st.st_mode)) {
			blocktype = '3';
			dev = st.st_rdev;
		} else if (S_ISBLK(st.st_mode)) {
			blocktype = '4';
			dev = st.st_rdev;
		} else if (S_ISFIFO(st.st_mode)) {
			blocktype = '6';
		} else {
			fprintf(stderr, "%s: %s: invalid file type\n", progname, pi->name);
			return -1;
		}
	}
	if (uname) {
		userptr = uname;
	} else {
		pw = getpwuid(st.st_uid);
		if (pw) {
			snprintf(username, 33, "%s", pw->pw_name);
			userptr = username;
		} else {
			fprintf(stderr, "%s: %s: warning: invalid UID %d\n", progname,
				pi->name, st.st_uid);
			userptr = NULL;
		}
		uid = st.st_uid;
	}
	if (gname) {
		groupptr = gname;
	} else {
		gr = getgrgid(st.st_gid);
		if (gr) {
			snprintf(groupname, 33, "%s", gr->gr_name);
			groupptr = groupname;
		} else {
			fprintf(stderr, "%s: %s: warning: invalid GID %d\n", progname,
				pi->name, st.st_gid);
			groupptr = NULL;
		}
		gid = st.st_gid;
	}
	if (path_is_untracked(pathbuf, untracked)) {
		switch (blocktype) {
		case '0':
			blocktype = 'O';
			break;
		case '1':
			blocktype = 'l';
			break;
		case '2':
			blocktype = 'Z';
			break;
		case '3':
			blocktype = 'z';
			break;
		case '4':
			blocktype = 'y';
			break;
		case '5':
			blocktype = 's';
			break;
		case '6':
			blocktype = 'b';
			break;
		}
	}
	if (blocktype == '0' || blocktype == 'O') {
		in = fopen(pi->name, "r");
		if (!in)
			return -1;
	}
	output_infoblock(out, outfn, pathbuf, st.st_mode & ~S_IFMT, uid, gid, size,
		st.st_mtime, blocktype, target, userptr, groupptr, dev);
	if (blocktype == '0' || blocktype == 'O') {
		fill_bytes(out, outfn,
			(512 - x_copy_all(in, pi->name, out, outfn) % 512) % 512);
		if (fclose(in)) {
			error(pi->name);
			return -1;
		}
	}
	return 0;
}

void output_footer(FILE *out, char *filename) {
	fill_bytes(out, filename, 1024);
}

int checksum_is_valid(char *buf, char *filename) {
	size_t i;
	int csum, scsum, dcsum;
	csum = checksum(buf, 0);
	scsum = checksum(buf, 1);
	dcsum = strtol(buf + 148, NULL, 8);
	if (dcsum != csum && dcsum != scsum) {
		fprintf(stderr,
			"%s: %s: bad checksum (expected %06o or %06o, found %.6s)\n",
			progname, filename, csum, scsum, buf + 148);
		return 0;
	}
	return 1;
}

/* 1: success, 0: no more headers */
int read_header(FILE *in, char *infn, char **filename, mode_t *mode,
		off_t *size, char *block_type, time_t *modtime, char **target,
		dev_t *dev) {
	char buf[512];
	char path[257];
	size_t len, fsize, tarsize;
	int devmajor, devminor;
	len = xread(in, infn, buf, 512);
	if (len == 0)
		return 0;
	if (len < 512) {
		fprintf(stderr, "%s: %s: short read\n", progname, infn);
		exit(1);
	}
	if (!buf[0])
		return 0;
	if (!checksum_is_valid(buf, infn))
		exit(1);
	memcpy(path, buf + 345, 155);
	len = strlen(path);
	if (len && path[len - 1] != '/') {
		path[len] = '/';
		len++;
	}
	memcpy(path + len, buf, 100);
	path[len + 100] = 0;
	len = strlen(path);
	if (!strcmp(path, "././@LongLink")) {
		fsize = strtol(buf + 124, NULL, 8);
		tarsize = ((fsize + 511) & ~511);
		*filename = xmalloc(tarsize + 1, infn);
		if (xread(in, infn, *filename, tarsize) < tarsize) {
			free(*filename);
			return 0;
		}
		(*filename)[tarsize] = 0;
		if (xread(in, infn, buf, 512) < 512) {
			free(*filename);
			return 0;
		}
		if (!buf[0]) {
			free(*filename);
			return 0;
		}
		if (!checksum_is_valid(buf, infn))
			exit(1);
	} else if (!path_is_safe(path)) {
		fprintf(stderr, "%s: %s: invalid path\n", progname, path);
		exit(1);
	} else {
		*filename = strdup(path);
		if (!*filename) {
			fprintf(stderr, "%s: %s: out of memory\n", progname, infn);
			exit(1);
		}
	}
	*size = strtol(buf + 124, NULL, 8);
	if (mode)
		*mode = strtol(buf + 100, NULL, 8);
	if (block_type)
		*block_type = buf[156];
	if (modtime)
		*modtime = strtol(buf + 136, NULL, 8);
	if (target) {
		*target = xmalloc(101, infn);
		memcpy(*target, buf + 157, 100);
		(*target)[100] = 0;
	}
	if (dev) {
		devmajor = strtol(buf + 329, NULL, 8);
		devminor = strtol(buf + 337, NULL, 8);
		*dev = makedev(devmajor, devminor);
	}
	return 1;
}

void trim_dirname(char *name) {
	size_t len;
	len = strlen(name);
	if (len > 1 && name[len - 1] == '/') {
		name[len - 1] = 0;
		len--;
	}
	for (; len >= 2 && !strcmp(name + len - 2, "/."); len -= 2)
		name[len - 2] = 0;
}

int replace_open(char *filename, mode_t mode) {
	struct stat st;
	if (delete(filename, 0))
		return -1;
	else
		return open(filename, O_WRONLY|O_CREAT|O_NOFOLLOW|O_TRUNC, mode);
}

int ex_write(FILE *in, char *infn, char *filename, mode_t mode, time_t modtime,
		size_t size, int reopen, int flags, char *tgtdir) {
	char buf[1024];
	struct timeval tv[2];
	struct stat st;
	FILE *out, *f;
	char *outfn, *ffn;
	mode_t um;
	size_t tdlen, fnlen, i, rs, cs, cps;
	int r, fd, oflags;
	r = 2;
	if (flags & EXTRACT_PERMISSION)
		um = umask(0);
	if (reopen) {
		fd = open(filename, O_WRONLY|O_NOFOLLOW);
	} else {
		oflags = O_WRONLY|O_CREAT|O_NOFOLLOW|O_TRUNC;
		if (flags & EXTRACT_KEEP)
			oflags |= O_EXCL;
		fd = open(filename, oflags, mode);
		if (fd < 0) {
			switch (errno) {
			case ENOENT:
				if (flags & EXTRACT_PERMISSION)
					umask(um);
				return 0;
			case ELOOP: case EISDIR: case EEXIST:
				if ((flags & EXTRACT_KEEP) && (flags & EXTRACT_NOISYKEEP))
					goto err0;
				if (!(flags & EXTRACT_KEEP))
					fd = replace_open(filename, mode);
				else
					return 1;
				break;
			case EACCES:
				if (flags & EXTRACT_KEEP)
					goto err0;
				else
					fd = replace_open(filename, mode);
				break;
			}
		}
	}
	if (fd < 0)
		goto err0;
	f = fdopen(fd, "w");
	if (!f) {
		close(fd);
		goto err0;
	}
	if (!tgtdir) {
		ffn = filename;
	} else {
		tdlen = strlen(tgtdir);
		fnlen = strlen(filename);
		if (tgtdir[tdlen - 1] == '/') {
			ffn = malloc(tdlen + fnlen + 1);
			if (!ffn) {
				fprintf(stderr, "%s: copying to %s%s: out of memory\n",
					progname, tgtdir, filename);
				exit(1);
			}
			sprintf(ffn, "%s%s", tgtdir, filename);
		} else {
			ffn = malloc(tdlen + fnlen + 2);
			if (!ffn) {
				fprintf(stderr, "%s: copying to %s/%s: out of memory\n",
					progname, tgtdir, filename);
				exit(1);
			}
			sprintf(ffn, "%s/%s", tgtdir, filename);
		}
	}
	cps = size;
	for (i = 0; i < size; i += 1024) {
		if (i + 1024 < size)
			cs = 1024;
		else
			cs = size - i;
		rs = fread(buf, 1, cs, in);
		if (ferror(in)) {
			fprintf(stderr, "%s: copying to %s: %s\n", progname, ffn,
				strerror(errno));
			exit(1);
		}
		if (rs)
			fwrite(buf, 1, rs, f);
		if (rs < cs) {
			cps = i + rs;
			break;
		}
	}
	if (cps < size) {
		fprintf(stderr, "%s: copying to %s: short read\n", progname,
			ffn);
		exit(1);
	}
	if (ferror(f)) {
		fprintf(stderr, "%s: copying to %s: %s\n", progname, ffn,
			strerror(errno));
		r = -1;
	}
	/* modification time and setuid changes below get overwritten if data is
		written to the file. */
	if (fflush(f)) {
		fprintf(stderr, "%s: copying to %s: %s\n", progname, ffn,
			strerror(errno));
		r = -1;
	}
	tv[0].tv_sec = time(NULL);
	tv[0].tv_usec = 0;
	tv[1].tv_sec = modtime;
	tv[1].tv_usec = 0;
	if (r >= 1 && (futimes(fd, tv) || ((flags & EXTRACT_PERMISSION) &&
			(mode & 07000) && fchmod(fd, mode) && errno != EPERM))) {
		fprintf(stderr,
			"%s: copying to %s: "
				"setting permissions and modification time: %s\n",
			progname, ffn, strerror(errno));
		r = -1;
	}
	if (flags & EXTRACT_PERMISSION)
		umask(um);
	if (fclose(f)) {
		fprintf(stderr, "%s: copying to %s: %s\n", progname, ffn,
			strerror(errno));
		r = -1;
	}
	if (ffn != filename)
		free(ffn);
	return r;
err0:
	fprintf(stderr, "%s: copying to %s: %s\n", progname, ffn, strerror(errno));
err:
	if (flags & EXTRACT_PERMISSION)
		umask(um);
	xseek(in, infn, size);
	if (ffn != filename)
		free(ffn);
	return -1;
}

int ex_symlink(char *target, char *filename, int flags, char *tgtdir) {
	int r;
	r = 2;
	errno = 0;
	if (symlink(target, filename)) {
		switch (errno) {
		case ENOENT:
			return 0;
		case EEXIST:
			if ((flags & EXTRACT_KEEP) && (flags & EXTRACT_NOISYKEEP))
				goto err;
			errno = 0;
			if (!(flags & EXTRACT_KEEP)) {
				delete(filename, 0);
				symlink(target, filename);
			} else {
				r = 1;
			}
			break;
		}
	}
	if (errno)
		goto err;
	return r;
err:
	if (!tgtdir)
		fprintf(stderr, "%s: creating symlink from %s to %s: %s\n", progname,
			filename, target, strerror(errno));
	else if (tgtdir[strlen(tgtdir) - 1] == '/')
		fprintf(stderr, "%s: creating symlink from %s%s to %s%s: %s\n",
			progname, tgtdir, filename, tgtdir, target, strerror(errno));
	else
		fprintf(stderr, "%s: creating symlink from %s/%s to %s/%s: %s\n",
			progname, tgtdir, filename, tgtdir, target, strerror(errno));
	return -1;
}

int ex_link(char *target, char *filename, int flags, char *tgtdir) {
	struct stat st;
	int r;
	r = 2;
	errno = 0;
	if (link(target, filename)) {
		switch (errno) {
		case ENOENT:
			if (stat(filename, &st))
				goto err;
			return 0;
		case EEXIST:
			if ((flags & EXTRACT_KEEP) && (flags & EXTRACT_NOISYKEEP))
				goto err;
			errno = 0;
			if (!(flags & EXTRACT_KEEP)) {
				delete(filename, 0);
				link(target, filename);
			} else {
				r = 1;
			}
			break;
		}
	}
	if (errno)
		goto err;
	return r;
err:
	/* the first entry in a TAR file for a file with several hard links must
		be marked with the actual file type, and the others must be marked hard
		hard link, so an attempt to create a hard link to a nonexistent file
		while extracting files in order represents a bad TAR file. */
	if (!tgtdir)
		fprintf(stderr, "%s: creating hard link from %s to %s: %s\n", progname,
			filename, target, strerror(errno));
	else if (tgtdir[strlen(tgtdir) - 1] == '/')
		fprintf(stderr, "%s: creating hard link from %s%s to %s%s: %s\n",
			progname, tgtdir, filename, tgtdir, target, strerror(errno));
	else
		fprintf(stderr, "%s: creating hard link from %s/%s to %s/%s: %s\n",
			progname, tgtdir, filename, tgtdir, target, strerror(errno));
	return -1;
}

int ex_mknod(char *filename, mode_t mode, dev_t dev, int flags, char *tgtdir) {
	mode_t um;
	int r;
	r = 2;
	errno = 0;
	if (flags & EXTRACT_PERMISSION)
		um = umask(0);
	if (mknod(filename, mode, dev)) {
		switch (errno) {
		case ENOENT:
			if (flags & EXTRACT_PERMISSION)
				umask(um);
			return 0;
		case EEXIST:
			if ((flags & EXTRACT_KEEP) && (flags & EXTRACT_NOISYKEEP))
				goto err;
			errno = 0;
			if (!(flags & EXTRACT_KEEP)) {
				delete(filename, 0);
				mknod(filename, mode, dev);
			} else {
				r = 1;
			}
			break;
		}
	}
	if (errno)
		goto err;
	if (flags & EXTRACT_PERMISSION)
		umask(um);
	return r;
err:
	if (flags & EXTRACT_PERMISSION)
		umask(um);
	if (!tgtdir)
		fprintf(stderr, "%s: creating special file %s: %s\n", progname,
			filename, strerror(errno));
	else if (tgtdir[strlen(tgtdir) - 1] == '/')
		fprintf(stderr, "%s: creating special file %s%s: %s\n", progname,
			tgtdir, filename, strerror(errno));
	else
		fprintf(stderr, "%s: creating special file %s/%s: %s\n", progname,
			tgtdir, filename, strerror(errno));
	return -1;
}

int ex_mkdir(char *filename, mode_t mode, int flags, char *tgtdir) {
	struct stat st;
	int r;
	mode_t um;
	errno = 0;
	r = 2;
	if (flags & EXTRACT_PERMISSION)
		um = umask(0);
	if (mkdir(filename, mode)) {
		switch (errno) {
		case ENOENT: case ENOTDIR:
			if (flags & EXTRACT_PERMISSION)
				umask(um);
			return 0;
		case EEXIST:
			if ((flags & EXTRACT_KEEP) && (flags & EXTRACT_NOISYKEEP))
				goto err;
			errno = 0;
			if (!(flags & EXTRACT_KEEP)) {
				errno = 0;
				if (lstat(filename, &st))
					goto err;
				if (S_ISDIR(st.st_mode)) {
					chmod(filename, mode);
				} else {
					delete(filename, 0);
					mkdir(filename, mode);
				}
			} else {
				r = 1;
			}
			break;
		}
	}
	if (errno)
		goto err;
	if (flags & EXTRACT_PERMISSION)
		umask(um);
	return r;
err:
	if (flags & EXTRACT_PERMISSION)
		umask(um);
	if (!tgtdir)
		fprintf(stderr, "%s: creating directory %s: %s\n", progname, filename,
			strerror(errno));
	else if (tgtdir[strlen(tgtdir) - 1] == '/')
		fprintf(stderr, "%s: creating directory %s%s: %s\n", progname,
			tgtdir, filename, strerror(errno));
	else
		fprintf(stderr, "%s: creating directory %s/%s: %s\n", progname,
			tgtdir, filename, strerror(errno));
	return -1;
}

char *backup_dir(char *tgtdir, char *name) {
	DIR *dir;
	struct dirent *de;
	char *tmp, *tmpdir, *src, *dest;
	size_t len, clen;
	len = strlen(name);
	tmp = malloc(len + 13);
	if (!tmp) {
		fprintf(stderr, "%s: %s: out of memory\n", progname, name);
		return NULL;
	}
	sprintf(tmp, "%s-tspm.XXXXXX", name);
	tmpdir = mkdtemp(tmp);
	if (!tmpdir) {
		target_error(tgtdir, name);
		goto err;
	}
	dir = opendir(name);
	if (!dir) {
		target_error(tgtdir, name);
		goto err;
	}
	for (de = readdir(dir); de; de = readdir(dir)) {
		if (strcmp(de->d_name, ".") && strcmp(de->d_name, "..")) {
			clen = strlen(de->d_name);
			src = malloc(len + clen + 2);
			if (!src) {
				target_emsg(tgtdir, name, "out of memory");
				goto err2;
			}
			dest = malloc(len + clen + 14);
			if (!dest) {
				target_emsg(tgtdir, name, "out of memory");
				goto err3;
			}
			sprintf(src, "%s/%s", name, de->d_name);
			sprintf(dest, "%s/%s", tmpdir, de->d_name);
			if (move(src, dest)) {
				target_error(tgtdir, src);
				goto err4;
			}
		}
	}
	closedir(dir);
	return tmpdir;
err4:
	free(dest);
err3:
	free(src);
err2:
	closedir(dir);
err:
	free(tmp);
	return NULL;
}

char *backup_file(char *tgtdir, char *filename) {
	char *tmp;
	int fd;
	tmp = malloc(strlen(filename) + 13);
	if (!tmp) {
		target_emsg(tgtdir, filename, "out of memory\n");
		return NULL;
	}
	sprintf(tmp, "%s-tspm.XXXXXX", filename);
	fd = mkstemp(tmp);
	if (fd < 0) {
		target_error(tgtdir, filename);
		free(tmp);
		return NULL;
	}
	if (close(fd))
		target_error(tgtdir, filename);
	errno = 0;
	if (move(filename, tmp)) {
		target_error(tgtdir, filename);
		delete(tmp, 0);
		free(tmp);
		return NULL;
	}
	return tmp;
}

struct file_list *file_list_add(struct file_list *fl, int dir, char *tgtdir,
		char *filename) {
	struct file_list *new, *curr;
	for (curr = fl; curr; curr = curr->next) {
		switch (paths_equal(filename, curr->filename)) {
		case 1:
			return fl;
		case -1:
			target_emsg(tgtdir, filename, "out of memory");
			return NULL;
		}
	}
	new = malloc(sizeof(struct file_list));
	if (!new) {
		target_emsg(tgtdir, filename, "out of memory");
		return NULL;
	}
	new->filename = strdup(filename);
	if (!(new->filename)) {
		target_emsg(tgtdir, filename, "out of memory");
		free(new);
		return NULL;
	}
	new->backup = NULL;
	new->dir = dir;
	new->next = fl;
	return new;
}

int add_filename(char *tgtdir, struct file_list *oldfl,
		struct file_list **newfl, int overwrite, int dir, char *filename,
		char *backup) {
	struct file_list *fl;
	fl = file_list_add(*newfl, dir, tgtdir, filename);
	if (!fl) {
		if (overwrite) {
			if (backup) {
				if (move(backup, filename))
					target_error(tgtdir, filename);
			} else {
				if (delete(filename, 1))
					target_error(tgtdir, filename);
			}
		}
		return -1;
	}
	*newfl = fl;
	if (overwrite) {
		for (fl = oldfl; fl; fl = fl->next) {
			switch (paths_equal(fl->filename, filename)) {
			case 1:
				fl->backup = backup;
				break;
			case -1:
				target_emsg(tgtdir, filename, "out of memory");
				return -1;
			}
		}
	}
	return 0;
}

char *backup(char *tgtdir, struct file_list *oldfl, char *filename) {
	struct file_list *fl;
	struct stat st;
	for (fl = oldfl; fl; fl = fl->next) {
		switch (paths_equal(fl->filename, filename)) {
		case 1:
			if (lstat(filename, &st))
				return NULL;
			if (S_ISDIR(st.st_mode))
				return backup_dir(tgtdir, filename);
			else
				return backup_file(tgtdir, filename);
		case -1:
			target_emsg(tgtdir, filename, "out of memory");
			return NULL;
		}
	}
	return NULL;
}

char *backup_or_fail(char *tgtdir, struct file_list *oldfl, int overwrite,
		char *filename) {
	char *tmp;
	if (overwrite) {
		tmp = backup(tgtdir, oldfl, filename);
		if (!tmp)
			return NULL;
		return tmp;
	}
	return NULL;
}

int tar_extractor_flags(int flags, int force, int overwrite, int tracking,
		int track) {
	if (tracking) {
		flags = EXTRACT_PERMISSION;
		if (!track || !force || overwrite)
			flags |= EXTRACT_KEEP;
		if (!force && !overwrite)
			flags |= EXTRACT_NOISYKEEP;
	}
	return flags;
}

int tar_extractor_write(char *tgtdir, struct file_list *oldfl,
		struct file_list **newfl, int flags, int force, int overwrite,
		int tracking, FILE *in, char *infn, char *filename, mode_t mode,
		time_t modtime, size_t size, int track, int reopen) {
	char *tmp;
	int r;
	tmp = NULL;
	r = ex_write(in, infn, filename, mode, modtime, size, reopen,
		tar_extractor_flags(flags, force, overwrite, tracking, track),
		tgtdir);
	if (r < 1)
		return r;
	if (r == 1) {
		if (track && tracking) {
			tmp = backup_or_fail(tgtdir, oldfl, overwrite, filename);
			if (tmp || force) {
				r = ex_write(in, infn, filename, mode, modtime, size, reopen,
					EXTRACT_PERMISSION, tgtdir);
			} else {
				target_emsg(tgtdir, filename, "File already exists");
				xseek(in, infn, size);
				return -1;
			}
		} else {
			xseek(in, infn, size);
		}
	}
	if (r < 1)
		return r;
	if (track && tracking &&
			add_filename(tgtdir, oldfl, newfl, overwrite, 0, filename, tmp)) {
		if (r > 1 && !overwrite)
			delete(filename, 0);
		return -1;
	}
	return 1;
}

int tar_extractor_symlink(char *tgtdir, struct file_list *oldfl,
		struct file_list **newfl, int flags, int force, int overwrite,
		int tracking, char *target, char *filename, int track) {
	char *tmp;
	int r;
	r = ex_symlink(target, filename, tar_extractor_flags(flags, force,
		overwrite, tracking, track), tgtdir);
	if (r < 1)
		return r;
	if (r == 1 && track && tracking) {
		tmp = backup_or_fail(tgtdir, oldfl, overwrite, filename);
		if (tmp || force) {
			r = ex_symlink(target, filename, EXTRACT_PERMISSION, tgtdir);
		} else {
			target_emsg(tgtdir, filename, "File already exists");
			return -1;
		}
	}
	if (r < 1)
		return r;
	if (track && tracking && add_filename(tgtdir, oldfl, newfl,
			overwrite, 0, filename, tmp)) {
		if (r > 1 && !overwrite)
			delete(filename, 0);
		return -1;
	}
	return 1;
}

int tar_extractor_link(char *tgtdir, struct file_list *oldfl,
		struct file_list **newfl, int flags, int force, int overwrite,
		int tracking, char *target, char *filename, int track) {
	char *tmp;
	int r;
	r = ex_link(target, filename,
		tar_extractor_flags(flags, force, overwrite, tracking, track), tgtdir);
	if (r < 1)
		return r;
	if (r == 1 && track && tracking) {
		tmp = backup_or_fail(tgtdir, oldfl, overwrite, filename);
		if (tmp || force) {
			r = ex_link(target, filename, EXTRACT_PERMISSION, tgtdir);
		} else {
			target_emsg(tgtdir, filename, "File already exists");
			return -1;
		}
	}
	if (r < 1)
		return r;
	/* one can not make a hard link to a directory */
	if (track && tracking &&
			add_filename(tgtdir, oldfl, newfl, overwrite, 0, filename, tmp)) {
		if (r > 1 && !overwrite)
			delete(filename, 0);
		return -1;
	}
	return 1;
}

int tar_extractor_mknod(char *tgtdir, struct file_list *oldfl,
		struct file_list **newfl, int flags, int force, int overwrite,
		int tracking, char *filename, mode_t mode, dev_t dev, int track) {
	char *tmp;
	int r;
	tmp = NULL;
	r = ex_mknod(filename, mode, dev,
		tar_extractor_flags(flags, force, overwrite, tracking, track), tgtdir);
	if (r < 1)
		return r;
	if (r == 1 && track && tracking) {
		tmp = backup_or_fail(tgtdir, oldfl, overwrite, filename);
		if (tmp || force) {
			r = ex_mknod(filename, mode, dev, EXTRACT_PERMISSION, tgtdir);
		} else {
			target_emsg(tgtdir, filename, "File already exists");
			return -1;
		}
	}
	if (r < 1)
		return r;
	if (track && tracking &&
			add_filename(tgtdir, oldfl, newfl, overwrite, 0, filename, tmp)) {
		if (r > 1 && !overwrite)
			delete(filename, 0);
		return -1;
	}
	return 1;
}

int tar_extractor_mkdir(char *tgtdir, struct file_list *oldfl,
		struct file_list **newfl, int flags, int force, int overwrite,
		int tracking, char *filename, mode_t mode, int track) {
	int r;
	if (tracking) {
		flags = EXTRACT_PERMISSION;
		if (!track || !force)
			flags |= EXTRACT_KEEP;
	}
	r = ex_mkdir(filename, mode, flags, tgtdir);
	if (r < 1)
		return r;
	if (track && tracking &&
			add_filename(tgtdir, oldfl, newfl, overwrite, 1, filename, NULL)) {
		if (r > 1 && !overwrite)
			delete(filename, 1);
		return -1;
	}
	return 1;
}

int guarantee_dir(char *tgtdir, struct file_list *oldfl,
		struct file_list **newfl, int flags, int force, int overwrite,
		int tracking, char *dpytgtdir, char *filename, mode_t mode,
		int track) {
	trim_dirname(filename);
	switch (tar_extractor_mkdir(tgtdir, oldfl, newfl, flags, force, overwrite,
			tracking, filename, mode, track)) {
	case -1:
		return -1;
	case 0:
		if (guarantee_parent_dir(tgtdir, oldfl, newfl, flags, force, overwrite,
				tracking, dpytgtdir, filename, mode, track))
			return -1;
		return -(tar_extractor_mkdir(tgtdir, oldfl, newfl, flags, force,
			overwrite, tracking, filename, mode, track) != 1);
	case 1:
		return 0;
	}
}

int guarantee_parent_dir(char *tgtdir, struct file_list *oldfl,
		struct file_list **newfl, int flags, int force, int overwrite,
		int tracking, char *dpytgtdir, char *filename, mode_t mode,
		int track) {
	char *s, *p;
	int r;
	s = strdup(filename);
	if (!s) {
		target_emsg(dpytgtdir, filename, "out of memory");
		return -1;
	}
	p = dirname(s);
	if (strcmp(p, "."))
		r = guarantee_dir(tgtdir, oldfl, newfl, flags, force, overwrite,
			tracking, dpytgtdir, p, 0755, track);
	else
		r = 0;
	free(s);
	return r;
}

int extractor_write(char *tgtdir, struct file_list *oldfl,
		struct file_list **newfl, int flags, int force, int overwrite,
		int tracking, FILE *in, char *infn, char *dpytgtdir, char *filename,
		mode_t mode, time_t modtime, size_t size, int track, int reopen) {
	int r;
	r = tar_extractor_write(tgtdir, oldfl, newfl, flags, force, overwrite,
		tracking, in, infn, filename, mode, modtime, size, track, reopen);
	if (!r) {
		if (guarantee_parent_dir(tgtdir, oldfl, newfl, flags, force, overwrite,
				tracking, dpytgtdir, filename, mode, track)) {
			xseek(in, infn, size);
			return -1;
		}
		r = tar_extractor_write(tgtdir, oldfl, newfl, flags, force, overwrite,
			tracking, in, infn, filename, mode, modtime, size, track, reopen);
	}
	if (r == 1)
		return 0;
	return -1;
}

int extractor_link(char *tgtdir, struct file_list *oldfl,
		struct file_list **newfl, int flags, int force, int overwrite,
		int tracking, char *dpytgtdir, char *filename, char *target,
		int track) {
	int r;
	r = tar_extractor_link(tgtdir, oldfl, newfl, flags, force, overwrite,
		tracking, target, filename, track);
	if (!r) {
		if (guarantee_parent_dir(tgtdir, oldfl, newfl, flags, force, overwrite,
				tracking, dpytgtdir, filename, 0777, track))
			return -1;
		r = tar_extractor_link(tgtdir, oldfl, newfl, flags, force, overwrite,
			tracking, target, filename, track);
	}
	if (r < 1)
		return -1;
	return 0;
}

int extractor_symlink(char *tgtdir, struct file_list *oldfl,
		struct file_list **newfl, int flags, int force, int overwrite,
		int tracking, char *dpytgtdir, char *filename, char *target,
		int track) {
	int r;
	r = tar_extractor_symlink(tgtdir, oldfl, newfl, flags, force, overwrite,
		tracking, target, filename, track);
	if (!r) {
		if (guarantee_parent_dir(tgtdir, oldfl, newfl, flags, force, overwrite,
				tracking, dpytgtdir, filename, 0777, track))
			return -1;
		r = tar_extractor_symlink(tgtdir, oldfl, newfl, flags, force,
			overwrite, tracking, target, filename, track);
	}
	if (r < 1)
		return -1;
	return 0;
}

int extractor_mknod(char *tgtdir, struct file_list *oldfl,
		struct file_list **newfl, int flags, int force, int overwrite,
		int tracking, char *dpytgtdir, char *filename,
		dev_t dev, mode_t mode, int track) {
	int r;
	r = tar_extractor_mknod(tgtdir, oldfl, newfl, flags, force, overwrite,
		tracking, filename, mode, dev, track);
	if (!r) {
		if (guarantee_parent_dir(tgtdir, oldfl, newfl, flags, force, overwrite,
				tracking, dpytgtdir, filename, mode, track))
			return -1;
		r = tar_extractor_mknod(tgtdir, oldfl, newfl, flags, force, overwrite,
			tracking, filename, mode, dev, track);
	}
	if (r < 1)
		return -1;
	return 0;
}

int list_file(FILE *in, char *infn, int trackable) {
	char *filename, *target, *msg;
	off_t size;
	char block_type;
	int result;
	result = 1;
	if (!read_header(in, infn, &filename, NULL, &size, &block_type, NULL, NULL,
			NULL))
		return 0;
	if (!trackable && block_type && (block_type < '0' && block_type > '9') &&
			block_type != 'g' && block_type != 'x')
		goto uft;
	switch (block_type) {
	case '0': case 0: case 'O':
		printf("%s\n", filename);
		xseek(in, infn, (size + 511) & ~511);
		break;
	case '1': case 'l': case '2': case 'Z': case '3': case 'z':
	case '4': case 'y': case '5': case 's': case '6': case 'b':
		printf("%s\n", filename);
		break;
	case 'g':
	case 'x':
		fprintf(stderr, "%s: warning: PAX header ignored\n", progname);
		xseek(in, infn, (size + 511) & ~511);
		break;
	default:
		goto uft;
	}
	goto cleanup;
uft:
	fprintf(stderr, "%s: %s: unknown file type '%c'\n", progname, filename,
		block_type);
	result = -1;
	xseek(in, infn, (size + 511) & ~511);
cleanup:
	free(filename);
	return result;
}

int extract_file(char *tgtdir, struct file_list *oldfl,
		struct file_list **newfl, int flags, int force, int overwrite,
		int tracking, char *dpytgtdir, FILE *in, char *infn, int trackable) {
	char *filename, *target, *msg;
	mode_t mode;
	off_t size;
	time_t modtime;
	dev_t dev;
	int result, track;
	char block_type;
	result = 1;
	if (!read_header(in, infn, &filename, &mode, &size, &block_type, &modtime,
			&target, &dev))
		return 0;
	track = !block_type || (block_type >= '0' && block_type <= '9') ||
		block_type == 'g' || block_type == 'x';
	if (!track && !trackable)
		goto uft;
	switch (block_type) {
	case '0': case 0: case 'O':
		if (extractor_write(tgtdir, oldfl, newfl, flags, force, overwrite,
				tracking, in, infn, dpytgtdir, filename, mode, modtime, size,
				track, 0))
			result = -1;
		xseek(in, infn, (512 - size % 512) % 512);
		break;
	case '1': case 'l':
		if (extractor_link(tgtdir, oldfl, newfl, flags, force, overwrite,
				tracking, dpytgtdir, filename, target, track))
			result = -1;
		break;
	case '2': case 'Z':
		if (extractor_symlink(tgtdir, oldfl, newfl, flags, force, overwrite,
				tracking, dpytgtdir, filename, target, track))
			result = -1;
		break;
	case '3': case 'z':
		if (extractor_mknod(tgtdir, oldfl, newfl, flags, force, overwrite,
				tracking, dpytgtdir, filename, dev, S_IFCHR | mode, track))
			result = -1;
		break;
	case '4': case 'y':
		if (extractor_mknod(tgtdir, oldfl, newfl, flags, force, overwrite,
				tracking, dpytgtdir, filename, dev, S_IFBLK | mode, track))
			result = -1;
		break;
	case '5': case 's':
		if (guarantee_dir(tgtdir, oldfl, newfl, flags, force, overwrite,
				tracking, dpytgtdir, filename, mode, track))
			result = -1;
		break;
	case '6': case 'b':
		if (extractor_mknod(tgtdir, oldfl, newfl, flags, force, overwrite,
				tracking, dpytgtdir, filename, dev, S_IFIFO | mode, track))
			result = -1;
		break;
	case 'g':
	case 'x':
		fprintf(stderr, "%s: warning: PAX header ignored\n", progname);
		xseek(in, infn, size + (512 - size % 512) % 512);
		break;
	default:
		goto uft;
	}
	goto cleanup;
uft:
	msg = malloc(22);
	if (msg) {
		sprintf(msg, "unknown file type '%c'", block_type);
		target_emsg(dpytgtdir, filename, msg);
		free(msg);
	} else {
		target_emsg(dpytgtdir, filename, "unknown file type");
	}
	result = -1;
	xseek(in, infn, size + (512 - size % 512) % 512);
cleanup:
	free(filename);
	free(target);
	return result;
}
