#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <sys/stat.h>
#include <sys/sysmacros.h>

#include "lib.h"
#include "fs.h"
#include "archive.h"

/* 0: success, -1: failure */
void cpio_output_header(FILE *out, char *filename, char *name, ino_t inode,
		mode_t mode, uid_t uid, gid_t gid, int nlink, int mtime, off_t size,
		dev_t dev, dev_t rdev) {
	unsigned char header[111];
	size_t pathlen;
	pathlen = strlen(name);
	snprintf(header, 7, "070701");
	snprintf(header + 6, 9, "%.8X", inode);
	snprintf(header + 14, 9, "%.8X", mode);
	snprintf(header + 22, 9, "%.8X", uid);
	snprintf(header + 30, 9, "%.8X", gid);
	snprintf(header + 38, 9, "%.8X", nlink);
	snprintf(header + 46, 9, "%.8X", mtime);
	snprintf(header + 54, 9, "%.8X", size);
	snprintf(header + 62, 9, "%.8X", major(dev));
	snprintf(header + 70, 9, "%.8X", minor(dev));
	snprintf(header + 78, 9, "%.8X", major(rdev));
	snprintf(header + 86, 9, "%.8X", minor(rdev));
	snprintf(header + 94, 9, "%.8X", pathlen + 1);
	snprintf(header + 102, 9, "00000000");
	xwrite(out, filename, header, 110);
	xwrite(out, filename, name, pathlen);
	fill_bytes(out, filename, 4 - (pathlen + 2) % 4);
}


/* 0: success, -1: failure */
int cpio_output_path(FILE *out, char *filename, struct path_info *pi,
		int replace_user, uid_t uid, int replace_group, gid_t gid) {
	FILE *in;
	struct stat st;
	off_t size;
	char *target, *userptr, *groupptr;
	if (lstat(pi->name, &st))
		goto err;
	if (!pi->link) {
		if (S_ISREG(st.st_mode)) {
			in = fopen(pi->name, "r");
			if (!in)
				return -1;
		} else if (S_ISLNK(st.st_mode)) {
			target = malloc(st.st_size);
			if (!target) {
				fprintf(stderr, "%s: %s: out of memory\n", progname, pi->name);
				return -1;
			}
			if (readlink(pi->name, target, st.st_size) < 0) {
				free(target);
				goto err;
			}
		}
	}
	if (!replace_user)
		uid = st.st_uid;
	if (!replace_group)
		gid = st.st_gid;
	if ((S_ISREG(st.st_mode) || S_ISLNK(st.st_mode)) && !pi->link)
		size = st.st_size;
	else
		size = 0;
	cpio_output_header(out, filename, pi->name, st.st_ino, st.st_mode,
		uid, gid, st.st_nlink, st.st_mtime, size, st.st_dev, st.st_rdev);
	if (!pi->link) {
		if (S_ISREG(st.st_mode)) {
			fill_bytes(out, filename,
				(4 - x_copy_all(in, pi->name, out, filename) % 4) % 4);
			if (fclose(in))
				goto err;
		} else if (S_ISLNK(st.st_mode)) {
			xwrite(out, filename, target, st.st_size);
			fill_bytes(out, filename, (4 - st.st_size % 4) % 4);
			free(target);
		}
	}
	return 0;
err:
	error(pi->name);
	return -1;
}

void cpio_output_footer(FILE *f, char *filename) {
	cpio_output_header(f, filename, "TRAILER!!!", 0, 0, 0, 0, 1, 0, 0, 0, 0);
}

int cpio_read_header(FILE *in, char *infn, char **filename, mode_t *mode,
		off_t *size, time_t *modtime, char **target, ino_t *inode,
		dev_t *dev, dev_t *rdev) {
	char buf[110];
	char numbuf[9];
	off_t namesize, namebufsize, tgtbufsize;
	int devmajor, devminor;
	numbuf[8] = 0;
	if (xread(in, infn, buf, 110) < 110)
		return 0;
	if (strncmp(buf, "070701", 6)) {
		fprintf(stderr, "%s: invalid format identifier\n", progname);
		exit(1);
	}
	if (inode) {
		memcpy(numbuf, buf + 6, 8);
		*inode = strtol(numbuf, NULL, 16);
	}
	if (mode) {
		memcpy(numbuf, buf + 14, 8);
		*mode = strtol(numbuf, NULL, 16);
	}
	if (modtime) {
		memcpy(numbuf, buf + 46, 8);
		*modtime = strtol(numbuf, NULL, 16);
	}
	if (dev) {
		memcpy(numbuf, buf + 62, 8);
		devmajor = strtol(numbuf, NULL, 16);
		memcpy(numbuf, buf + 70, 8);
		devminor = strtol(numbuf, NULL, 16);
		*dev = makedev(devmajor, devminor);
	}
	if (rdev) {
		memcpy(numbuf, buf + 78, 8);
		devmajor = strtol(numbuf, NULL, 16);
		memcpy(numbuf, buf + 86, 8);
		devminor = strtol(numbuf, NULL, 16);
		*rdev = makedev(devmajor, devminor);
	}
	memcpy(numbuf, buf + 54, 8);
	*size = strtol(numbuf, NULL, 16);
	memcpy(numbuf, buf + 94, 8);
	namesize = strtol(numbuf, NULL, 16);
	namebufsize = namesize + (4 - (namesize + 2) % 4) % 4;
	*filename = malloc(namebufsize);
	if (!*filename) {
		fprintf(stderr, "%s: out of memory\n", progname);
		exit(1);
	}
	if (xread(in, infn, *filename, namebufsize) < namebufsize) {
		fprintf(stderr, "%s: short read\n", progname);
		return -1;
	}
	if (!path_is_safe(*filename)) {
		fprintf(stderr, "%s: %s: invalid path\n", progname, *filename);
		exit(1);
	}
	if (S_ISLNK(*mode)) {
		tgtbufsize = *size + (4 - *size % 4) % 4;
		if (target) {
			*target = malloc(tgtbufsize + 1);
			if (!*target) {
				fprintf(stderr, "%s: out of memory\n", progname);
				exit(1);
			}
			if (xread(in, infn, *target, tgtbufsize) < tgtbufsize) {
				fprintf(stderr, "%s: short read\n", progname);
				return -1;
			}
			(*target)[tgtbufsize] = 0;
		} else {
			xseek(in, infn, tgtbufsize);
		}
	}
	return 1;
}

/* 1: file processed successfully, 0: no more files,
	-1: file processed with errors */
int cpio_list_file(FILE *in, char *infn, struct path_info **pi) {
	struct path_info **cpi;
	char *filename;
	ino_t inode;
	off_t size;
	mode_t mode;
	dev_t dev;
	int result, link;
	result = 1;
	link = 0;
	if (!cpio_read_header(in, infn, &filename, &mode, &size, NULL, NULL,
			&inode, &dev, NULL))
		return 0;
	for (cpi = pi; *cpi; cpi = &((*cpi)->next)) {
		if (inode == (*cpi)->inode && dev == (*cpi)->device) {
			printf("%s\n", filename);
			link = 1;
			goto found;
		}
	}
	*cpi = xmalloc(sizeof(struct path_info), filename);
	(*cpi)->name = strdup(filename);
	if (!(*cpi)->name) {
		free(*cpi);
		fprintf(stderr, "%s: %s: out of memory\n", progname, filename);
		exit(1);
	}
	(*cpi)->next = NULL;
	(*cpi)->inode = inode;
	(*cpi)->device = dev;
found:
	if (S_ISREG(mode)) {
		xseek(in, infn, (size + 3) / 4 * 4);
		printf("%s\n", filename);
	} else if (!link) {
		printf("%s\n", filename);
	}
cleanup:
	free(filename);
	return result;
}

/* 1: file processed successfully, 0: no more files,
	-1: file processed with errors */
int cpio_extract_file(char *tgtdir, struct file_list *oldfl,
		struct file_list **newfl, int flags, int force, int overwrite,
		int tracking, char *dpytgtdir, FILE *in, char *infn,
		struct path_info **pi) {
	struct path_info **cpi;
	FILE *f;
	char *filename, *target;
	ino_t inode;
	mode_t mode;
	off_t size;
	time_t modtime;
	dev_t dev, rdev;
	int result, link;
	result = 1;
	target = NULL;
	link = 0;
	if (!cpio_read_header(in, infn, &filename, &mode, &size, &modtime, &target,
			&inode, &dev, &rdev))
		return 0;
	for (cpi = pi; *cpi; cpi = &((*cpi)->next)) {
		if (inode == (*cpi)->inode && dev == (*cpi)->device) {
			if (extractor_link(tgtdir, oldfl, newfl, flags, force, overwrite,
					tracking, dpytgtdir, filename, (*cpi)->name, 0)) {
				result = -1;
				xseek(in, infn, size + (4 - size % 4) % 4);
				goto cleanup;
			}
			link = 1;
			goto found;
		}
	}
	*cpi = xmalloc(sizeof(struct path_info), filename);
	(*cpi)->name = strdup(filename);
	if (!(*cpi)->name) {
		free(*cpi);
		fprintf(stderr, "%s: %s: out of memory\n", progname, filename);
		exit(1);
	}
	(*cpi)->next = NULL;
	(*cpi)->inode = inode;
	(*cpi)->device = dev;
found:
	if (S_ISREG(mode)) {
		if (extractor_write(tgtdir, oldfl, newfl, flags, force, overwrite,
				tracking, in, infn, dpytgtdir, filename, mode, modtime, size,
				0, link))
			result = -1;
		xseek(in, infn, (4 - size % 4) % 4);
	} else if (S_ISLNK(mode)) {
		if (!link && extractor_symlink(tgtdir, oldfl, newfl, flags, force,
				overwrite, tracking, dpytgtdir, filename, target, 0))
			result = -1;
	} else if (S_ISDIR(mode)) {
		if (!link && guarantee_dir(tgtdir, oldfl, newfl, flags, force,
				overwrite, tracking, dpytgtdir, filename, mode, 0))
			result = -1;
	} else {
		if (!link && extractor_mknod(tgtdir, oldfl, newfl, flags, force,
				overwrite, tracking, dpytgtdir, filename, rdev, mode, 0))
			result = -1;
	}
cleanup:
	free(filename);
	if (target)
		free(target);
	return result;
}

void read_user_group(char *s, int *replace_user, uid_t *uid,
		int *replace_group, gid_t *gid) {
	struct passwd *pw;
	struct group *gr;
	char *t;
	t = strchr(s, ':');
	if (!t)
		t = strchr(s, '.');
	if (t) {
		*t = 0;
		t++;
	}
	if (*s) {
		pw = getpwnam(s);
		if (pw) {
			*replace_user = 1;
			*uid = pw->pw_uid;
		} else {
			fprintf(stderr, "%s: warning: invalid %s\n", progname, s);
		}
	}
	if (t && *t) {
		gr = getgrnam(t);
		if (gr) {
			*replace_group = 1;
			*gid = gr->gr_gid;
		} else {
			fprintf(stderr, "%s: warning: invalid %s\n", progname, t);
		}
	}
}

void usage() {
	fprintf(stderr,
		"\n%s -i [-R OWNER:GROUP] [-F FILENAME]\n"
		"\tExtract the archive on standard input or in FILENAME. Set owner\n"
		"\tof extracted files to OWNER if specified and group to GROUP if\n"
		"\tspecified.\n"
		"%s -t [-R OWNER:GROUP] [-F FILENAME]\n"
		"\tList the files in the archive on standard input or in FILENAME.\n"
		"\tSet owner of extracted files to OWNER if specified and group to\n"
		"\tGROUP if specified.\n"
		"%s -o [-R OWNER:GROUP] [-F FILENAME]\n"
		"\tCreate an archive on standard output or in FILENAME containing\n"
		"\tthe files listed on standard input. Set owner of extracted files\n"
		"\tto OWNER if specified and group to GROUP if specified.\n"
		"%s -e\n"
		"\tDisplay the version of this command.\n"
		"%s -?\n"
		"\tDisplay this usage message.\n\n",
		progname, progname, progname, progname, progname);
}

#define NONE 0
#define CREATE 1
#define LIST 2
#define EXTRACT 3

int main(int argc, char **argv) {
	int i, j, iinc, action, ch, oom, r, result, replace_user, replace_group;
	uid_t uid;
	gid_t gid;
	char *argfile, *tmp, *filename, *cpiofile;
	size_t arglen, pos;
	FILE *file;
	struct path_info *pi, *cpi;
	struct deferred_symlink *ds;
	progname = argv[0];
	filename = "-";
	action = NONE;
	replace_user = 0;
	uid = 0;
	gid = 0;
	for (i = 1; i < argc; i += iinc + 1) {
		iinc = 0;
		if (argv[i][0] == '-' && argv[i][1]) {
			for (j = 1; argv[i][j]; j++) {
				switch (argv[i][j]) {
				case 'e':
					version("cpio");
					exit(0);
				case 'F':
					if (i + iinc + 1 >= argc || argv[i+iinc+1][0] == '-') {
						fprintf(stderr,
							"%s: -F requires argument (%s -? for usage)\n",
							progname, progname);
						exit(1);
					}
					iinc++;
					filename = argv[i + iinc];
					break;
				case 'i':
					action = EXTRACT;
					break;
				case 'o':
					action = CREATE;
					break;
				case 'R':
					if (i + iinc + 1 >= argc || argv[i+iinc+1][0] == '-') {
						fprintf(stderr,
							"%s: -R requires argument (%s -? for usage)\n",
							progname, progname);
						exit(1);
					}
					iinc++;
					read_user_group(argv[i + iinc], &replace_user, &uid,
						&replace_group, &gid);
					break;
				case 't':
					action = LIST;
					break;
				case '?':
					version("cpio");
					usage();
					exit(0);
				default:
					fprintf(stderr, "%s: invalid -%c ('%s -?' for usage)\n",
						progname, argv[i][j], progname);
					return 1;
				}
			}
		}
	}
	result = 0;
	switch (action) {
	case NONE:
		fprintf(stderr, "%s: no -o, -i, or -t ('%s -?' for usage)\n", progname,
			progname);
		return 1;
	case CREATE:
		pi = NULL;
		arglen = 16;
		argfile = malloc(arglen);
		if (!argfile) {
			fprintf(stderr, "%s: out of memory\n");
			return 1;
		}
		pos = 0;
		oom = 0;
		for (ch = getchar(); ch >= 0; ch = getchar()) {
			if (ch == '\n') {
				if (!oom) {
					argfile[pos] = 0;
					cpiofile = strdup(argfile);
					if (cpiofile) {
						if (prepare_single_path_info(&pi, cpiofile, 1))
							result = 1;
					} else {
						fprintf(stderr, "%s: %s: out of memory\n", progname,
							argfile);
						result = 1;
					}
				}
				pos = 0;
				oom = 0;
			} else if (!oom) {
				argfile[pos] = ch;
				pos++;
				if (arglen == pos) {
					tmp = argfile;
					argfile = realloc(argfile, arglen * 2);
					if (!argfile) {
						fprintf(stderr, "%s: out of memory\n", progname);
						oom = 1;
						argfile = tmp;
						result = 1;
					}
					arglen *= 2;
				}
			}
		}
		if (pos && !oom) {
			argfile[pos] = 0;
			cpiofile = strdup(argfile);
			if (cpiofile) {
				if (prepare_single_path_info(&pi, cpiofile, 1))
					result = 1;
			} else {
				fprintf(stderr, "%s: %s: out of memory\n", progname,
					argfile);
				result = 1;
			}
		}
		file = filename_adjust_open(&filename, "w");
		if (!file)
			return 1;
		for (cpi = pi; cpi; cpi = cpi->next)
			if (cpio_output_path(file, filename, cpi, replace_user, uid,
					replace_group, gid))
				result = 1;
		cpio_output_footer(file, filename);
		if (fclose(file)) {
			error(filename);
			result = 1;
		}
		break;
	case EXTRACT: case LIST:
		file = filename_adjust_open(&filename, "r");
		pi = NULL;
		if (!file)
			return 1;
		if (action == LIST) {
			for (r = cpio_list_file(file, filename, &pi); r;
					r = cpio_list_file(file, filename, &pi))
				if (r < 0)
					result = 1;
		} else if (action == EXTRACT) {
			for (r = cpio_extract_file(NULL, NULL, NULL, 0, 0, 0, 0, NULL,
						file, filename, &pi); r;
					r = cpio_extract_file(NULL, NULL, NULL, 0, 0, 0, 0, NULL,
						file, filename, &pi))
				if (r < 0)
					result = 1;
		}
		if (fclose(file)) {
			error(filename);
			result = 1;
		}
		break;
	}
	return result;
}
