#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <libgen.h>
#include <sys/stat.h>

#include "lib.h"

/* 0: success, -1: error */
int canonical_basename_dirname(char *s, char **dir, char **base) {
	char *t, *u, *v, *w, *x;
	int steps;
	t = strdup(s);
	if (!t)
		return -1;
	u = basename(t);
	v = strdup(s);
	if (!v) {
		free(t);
		return -1;
	}
	w = dirname(v);
	steps = 0;
	while ((!strcmp(u, ".") && strcmp(w, ".")) || !strcmp(u, "..")) {
		if (!strcmp(u, ".."))
			steps++;
		free(t);
		t = strdup(w);
		if (!t) {
			free(v);
			return -1;
		}
		u = basename(t);
		x = strdup(w);
		free(v);
		if (!x) {
			free(t);
			return -1;
		}
		v = x;
		w = dirname(v);
	}
	for (; steps; steps--) {
		if (!strcmp(u, ".")) {
			free(t);
			t = strdup("..");
			if (!t) {
				free(v);
				return -1;
			}
			u = t;
		} else if (!strcmp(u, "..")) {
			x = malloc(strlen(w) + 4);
			if (!x) {
				free(t);
				free(v);
				return -1;
			}
			sprintf(x, "%s/..", v);
			free(v);
			v = x;
			w = x;
		} else if (strcmp(w, "/")) {
			free(t);
			t = strdup(w);
			if (!t) {
				free(v);
				return -1;
			}
			u = basename(t);
			x = strdup(w);
			free(v);
			if (!x) {
				free(t);
				return -1;
			}
			v = x;
			w = dirname(v);
		}
	}
	*dir = strdup(w);
	free(v);
	if (!*dir) {
		free(t);
		return -1;
	}
	*base = strdup(u);
	free(t);
	if (!*base)
		return -1;
	return 0;
}

int paths_equal(char *s, char *t) {
	char *u, *v, *sb, *sd, *tb, *td;
	int r;
	u = strdup(s);
	if (!u)
		return -1;
	v = strdup(t);
	if (!v) {
		free(u);
		return -1;
	}
	while (strcmp(u, ".") && strcmp(v, ".") &&
			strcmp(u, "/") && strcmp(v, "/")) {
		canonical_basename_dirname(u, &sd, &sb);
		canonical_basename_dirname(v, &td, &tb);
		r = strcmp(sb, tb);
		free(u);
		free(v);
		free(sb);
		free(tb);
		if (r) {
			free(sd);
			free(td);
			return 0;
		}
		u = sd;
		v = td;
	}
	r = strcmp(u, v);
	free(u);
	free(v);
	return !r;
}

char *construct_child_path(char *parent, char *child) {
	char *result;
	result = malloc(strlen(parent) + strlen(child) + 2);
	if (!result)
		return NULL;
	sprintf(result, "%s/%s", parent, child);
	return result;
}

int path_is_safe(char *path) {
	size_t len;
	len = strlen(path);
	return *path != '/' && !strstr(path, "/../") &&
		(len < 3 || strcmp("/..", path + len - 3)) && strcmp("..", path) &&
		strncmp("../", path, 3);
}

int delete(char *filename, int recursive) {
	DIR *dir;
	struct stat st;
	struct dirent *de;
	char *s;
	int result;
	result = 0;
	if (lstat(filename, &st))
		return -1;
	if (S_ISDIR(st.st_mode)) {
		if (recursive) {
			dir = opendir(filename);
			if (!dir)
				return -1;
			for (de = readdir(dir); de; de = readdir(dir)) {
				if (strcmp(de->d_name, ".") && strcmp(de->d_name, "..")) {
					s = malloc(strlen(filename) + strlen(de->d_name) + 2);
					if (!s) {
						errno = ENOMEM;
						return -1;
					}
					sprintf(s, "%s/%s", filename, de->d_name);
					if (delete(s, 1))
						result = -1;
					free(s);
				}
			}
		}
		if (rmdir(filename))
			result = -1;
		return result;
	} else {
		return unlink(filename);
	}
}

int copy(char *src, char *dest, int recursive) {
	DIR *dir;
	struct stat st;
	struct dirent *de;
	FILE *out, *inf;
	ssize_t rs;
	char path[PATH_MAX + 1];
	char *s, *t;
	int ofd, ifd, r;
	if (lstat(src, &st))
		return -1;
	if (S_ISDIR(st.st_mode)) {
		if (recursive) {
			if (!mkdir(dest, st.st_mode))
				return -1;
			r = 0;
			dir = opendir(src);
			if (!dir)
				return -1;
			for (de = readdir(dir); de; de = readdir(dir)) {
				if (strcmp(de->d_name, ".") && strcmp(de->d_name, "..")) {
					s = malloc(strlen(src) + strlen(de->d_name) + 2);
					if (!s) {
						errno = ENOMEM;
						return -1;
					}
					t = malloc(strlen(dest) + strlen(de->d_name) + 2);
					if (!t) {
						errno = ENOMEM;
						free(s);
						return -1;
					}
					sprintf(s, "%s/%s", src, de->d_name);
					sprintf(t, "%s/%s", dest, de->d_name);
					if (copy(s, t, 1))
						r = -1;
					free(t);
					free(s);
				}
			}
			return r;
		}
	} else if (S_ISREG(st.st_mode)) {
		ifd = open(src, O_RDONLY);
		if (ifd < 0)
			return -1;
		inf = fdopen(ifd, "r");
		if (!inf) {
			close(ifd);
			return -1;
		}
		ofd = open(dest, O_WRONLY|O_CREAT|O_TRUNC, st.st_mode);
		if (ofd < 0)
			return -1;
		out = fdopen(ofd, "w");
		if (!out) {
			close(ofd);
			return -1;
		}
		r = 0;
		copy_all(inf, src, out, dest);
		if (ferror(inf) || ferror(out))
			r = -1;
		if (fclose(out))
			r = -1;
		if (fclose(inf))
			r = -1;
		if (r)
			goto err;
		return 0;
	} else if (S_ISLNK(st.st_mode)) {
		rs = readlink(src, path, PATH_MAX);
		if (rs < 0)
			return -1;
		path[r] = 0;
		return symlink(path, dest);
	} else {
		return mknod(dest, st.st_mode, st.st_rdev);
	}
err:
	delete(dest, 1);
	return -1;
}

int move(char *src, char *dest) {
	if (rename(src, dest) &&
			(errno != EXDEV || copy(src, dest, 1) || delete(src, 1)))
		return -1;
	return 0;
}
