#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/stat.h>

#include "lib.h"
#include "fs.h"
#include "archive.h"
#include "package.h"

void overwrite_dependency_check(char *name, char *vers, char *depname,
		char *target) {
	char *datafile, *key, *value;
	FILE *f;
	int found;
	found = 0;
	if (strcmp(name, depname)) {
		f = open_package_datafile(depname, target, &datafile);
		if (f) {
			while (read_kvp(f, datafile, &key, &value)) {
				if (!strcmp(key, "Dependency")) {
					if (strcmp(value, name)) {
						if (found)
							goto cleanup;
					} else {
						found = 1;
					}
				} else if (!strcmp(key, "Minimum") && found) {
					if (strverscmp(vers, value) < 0) {
						fprintf(stderr,
							"%s: package %s depends on at least version %s of "
							"package %s (%s would be installed)\n", progname,
							depname, value, name, vers);
						exit(1);
					}
				} else if (!strcmp(key, "Maximum") && found) {
					if (strverscmp(vers, value) > 0) {
						fprintf(stderr,
							"%s: package %s depends on at most version %s of "
							"package %s (%s would be installed)\n", progname,
							depname, value, name, vers);
						exit(1);
					}
				}
			}
cleanup:
			cleanup_package_datafile(f, target, datafile);
		}
	}
}

void dependency_check(struct dep *dep, char *target) {
	char *datafile, *key, *value;
	FILE *f;
	int found;
	f = open_package_datafile(dep->name, target, &datafile);
	if (!f) {
		fprintf(stderr, "%s: dependency %s not installed\n", progname,
			dep->name);
		exit(1);
	}
	found = 0;
	while (read_kvp(f, datafile, &key, &value)) {
		if (!strcmp(key, "Version")) {
			if (dep->min_version && *(dep->min_version) &&
					strverscmp(value, dep->min_version) < 0) {
				fprintf(stderr,
					"%s: dependency %s must be at least version "
					"%s (%s installed)\n", progname, dep->name,
					dep->min_version, value);
				exit(1);
			}
			if (dep->max_version && *(dep->max_version) &&
					strverscmp(value, dep->max_version) > 0) {
				fprintf(stderr,
					"%s: dependency %s must be at most version "
					"%s (%s installed)\n", progname, dep->name,
					dep->max_version, value);
				exit(1);
			}
			cleanup_package_datafile(f, target, datafile);
			return;
			found = 1;
		}
		free(key);
		free(value);
		if (found)
			break;
	}
	cleanup_package_datafile(f, target, datafile);
	errno = 0;
	if (!found) {
		if ((dep->max_version && *(dep->max_version)) ||
				(dep->min_version && *(dep->min_version))) {
			fprintf(stderr,
				"%s: internal record store corrupted: can not determine "
				"version of dependency %s\n", progname, dep->name);
			exit(1);
		} else {
			fprintf(stderr,
				"%s: warning: internal record store corrupted: can not "
				"determine version of dependency %s\n", progname, dep->name);
		}
	}
}

int main(int argc, char **argv) {
	struct dirent *de;
	struct dep *deps, *cd;
	struct file_list *fl, *oldfl, *newfl;
	DIR *d;
	FILE *f, *df;
	char *name, *vers, *datafile, *filename, *key, *value, *target;
	int r, fd, i, j, overwrite, iinc, force;
	progname = argv[0];
	filename = NULL;
	target = "/";
	overwrite = 0;
	force = 0;
	for (i = 1; i < argc; i += iinc + 1) {
		iinc = 0;
		if (argv[i][0] == '-' && argv[i][1]) {
			for (j = 1; argv[i][j]; j++) {
				switch (argv[i][j]) {
				case 'f':
					force = 1;
					break;
				case 'o':
					overwrite = 1;
					break;
				case 't':
					if (i + iinc + 1 >= argc || argv[i+iinc+1][0] == '-') {
						fprintf(stderr, "%s: -t requires argument\n",
							progname);
						exit(1);
					}
					iinc++;
					target = argv[i + iinc];
					break;
				case 'V':
					version("inspak");
					return 0;
				case '?':
					version("inspak");
					fprintf(stderr,
						"\n%s [-of] [-t TARGET] FILENAME\n"
						"\tInstall a package from the package file named\n"
						"\tFILENAME. If -o is specified, replace any package\n"
						"\talready installed with the same name. If -f is\n"
						"\tspecified, overwrite any files already present\n"
						"\twith those from the package. If -t is specified,\n"
						"\ttreat TARGET as the root directory.\n"
						"%s -?\n"
						"\tShow this usage message.\n"
						"%s -V\n"
						"\tDisplay the version of this command.\n",
						progname, progname, progname);
					return 0;
				default:
					fprintf(stderr, "%s: invalid -%c\n", progname, argv[i][j]);
					exit(1);
				}
			}
		} else {
			if (filename) {
				fprintf(stderr,
					"%s: invalid %s: too many filenames on command line\n",
					progname, argv[i]);
				exit(1);
			}
			filename = argv[i];
		}
	}
	if (!filename) {
		fprintf(stderr, "%s: filename required\n", progname);
		exit(1);
	}
	f = xz_stream_open(filename, "r");
	if (!f)
		exit(1);
	read_package_header(f, filename, &name, &vers, &deps);
	if (chdir(target))
		die(target);
	if (overwrite) {
		d = opendir("etc/tspm");
		if (!d) {
			if (errno != ENOENT)
				target_die(target, "etc/tspm");
		} else {
			errno = 0;
			for (de = readdir(d); de; de = readdir(d)) {
				if (strcmp(de->d_name, ".") && strcmp(de->d_name, "..")) {
					overwrite_dependency_check(name, vers, de->d_name, target);
					errno = 0;
				}
			}
			if (errno)
				target_die(target, "etc/tspm");
			closedir(d);
		}
	}
	for (cd = deps; cd; cd = cd->next)
		dependency_check(cd, target);
	datafile = malloc(10 + strlen(name));
	if (!datafile) {
		fprintf(stderr, "%s: out of memory\n", progname);
		return 1;
	}
	sprintf(datafile, "etc/tspm/%s", name);
	errno = 0;
	fd = open(datafile, O_WRONLY|O_CREAT|O_NOFOLLOW|O_EXCL|O_TRUNC, 0644);
	if (fd < 0 && errno == ENOENT) {
		if (overwrite) {
			fprintf(stderr, "%s: package %s not installed\n", progname, name);
			exit(1);
		} else {
			errno = 0;
			r = mkdir("etc/tspm", 0755);
			if (r && errno == ENOENT) {
				if (mkdir("etc", 0755))
					target_die(target, "etc");
				r = mkdir("etc/tspm", 0755);
			}
			if (r)
				target_die(target, "etc/tspm");
			fd = open(datafile, O_WRONLY|O_CREAT|O_NOFOLLOW|O_EXCL|O_TRUNC,
				0644);
		}
	}
	oldfl = NULL;
	if (fd < 0) {
		if (errno == EEXIST) {
			if (overwrite) {
				df = fopen(datafile, "r");
				if (!f)
					target_die(target, datafile);
				while (read_kvp(df, datafile, &key, &value)) {
					if (!strcmp(key, "Directory")) {
						fl = file_list_add(oldfl, 1, target, value);
						if (!fl)
							exit(1);
						oldfl = fl;
					} else if (!strcmp(key, "File"))  {
						fl = file_list_add(oldfl, 0, target, value);
						if (!fl)
							exit(1);
						oldfl = fl;
					}
					free(key);
					free(value);
				}
				fd = open(datafile, O_WRONLY|O_NOFOLLOW|O_TRUNC, 0644);
				if (fd < 0)
					target_die(target, datafile);
			} else {
				fprintf(stderr, "%s: package %s already installed\n", progname,
					name);
				exit(1);
			}
		} else {
			target_die(target, datafile);
		}
	}
	df = fdopen(fd, "w");
	if (!df)
		target_die(target, datafile);
	newfl = NULL;
	for (r = extract_file(target, oldfl, &newfl, 0, force, overwrite, 1,
				target, f, filename, 1); r;
			r = extract_file(target, oldfl, &newfl, 0, force, overwrite, 1,
				target, f, filename, 1))
		if (r < 0)
			goto rbk;
	if (fclose(f))
		error(filename);
	if (write_kvp(df, "Version", vers))
		goto out_err;
	for (cd = deps; cd; cd = cd->next) {
		if (cd->name && *(cd->name)) {
			if (write_kvp(df, "Dependency", cd->name))
				goto out_err;
			if (cd->min_version && *(cd->min_version) &&
					write_kvp(df, "Minimum", cd->min_version))
				goto out_err;
			if (cd->max_version && *(cd->max_version) &&
					write_kvp(df, "Maximum", cd->max_version))
				goto out_err;
		}
	}
	fl = newfl;
	for (; fl; fl = fl->next) {
		if (fl->dir) {
			if (write_kvp(df, "Directory", fl->filename))
				goto out_err;
		} else {
			if (write_kvp(df, "File", fl->filename))
				goto out_err;
		}
	}
	if (fclose(df))
		target_error(target, datafile);
	errno = 0;
	if (overwrite && cleanup_overwrite(target, oldfl))
		return 1;
	return 0;
out_err:
	target_error(target, datafile);
rbk:
	if (overwrite) {
		if (rollback_for_overwrite(df, target, oldfl, newfl))
			target_error(target, datafile);
	} else {
		rollback(newfl);
	}
	if (fclose(df))
		target_error(target, datafile);
	if (!overwrite && delete(datafile, 0))
		target_error(target, datafile);
	return 1;
}
