#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

#include "lib.h"
#include "fs.h"

extern char **environ;

char *progname;

void *xmalloc(size_t size, char *filename) {
	void *result;
	result = malloc(size);
	if (!result) {
		fprintf(stderr, "%s: %s: out of memory\n", progname, filename);
		exit(1);
	}
	return result;
}

void *xrealloc(void *buf, size_t size, char *filename) {
	void *result;
	result = realloc(buf, size);
	if (!result) {
		fprintf(stderr, "%s: %s: out of memory\n", progname, filename);
		exit(1);
	}
	return result;
}

void error(char *filename) {
	fprintf(stderr, "%s: %s: %s\n", progname, filename, strerror(errno));
	errno = 0;
}

void die(char *filename) {
	error(filename);
	exit(1);
}

FILE *filename_adjust_open(char **filename, char *mode) {
	FILE *result;
	result = NULL;
	if (!strcmp(*filename, "-")) {
		if (!strcmp(mode, "w")) {
			result = stdout;
			*filename = "(standard output)";
		} else if (!strcmp(mode, "r")) {
			result = stdin;
			*filename = "(standard input)";
		}
	} else {
		result = fopen(*filename, mode);
		if (!result) {
			error(*filename);
			return NULL;
		}
	}
	return result;
}

FILE *compressed_stream_open(char *filename, char *mode,
		char *program, char *cmd) {
	FILE *f;
	char *argv[4];
	int pipefds[2];
	int fpart, outfd;
	if (pipe(pipefds)) {
		error(filename);
		return NULL;
	}
	argv[0] = cmd;
	if (*mode == 'r') {
		argv[1] = "-dc";
		argv[2] = filename;
	} else if (*mode == 'w') {
		argv[1] = "-";
		argv[2] = NULL;
		outfd = open(filename, O_WRONLY|O_CREAT|O_TRUNC, 0644);
		if (outfd < 0)
			goto err;
	}
	argv[3] = NULL;
	errno = 0;
	fpart = 0;
	if (!vfork()) {
		if (*mode == 'r') {
			if (close(1))
				_exit(1);
			if (dup2(pipefds[1], 1) < 0)
				_exit(1);
			if (close(pipefds[1]))
				_exit(1);
		} else if (*mode == 'w') {
			if (close(0))
				_exit(1);
			if (dup2(pipefds[0], 0) < 0)
				_exit(1);
			if (close(pipefds[0]))
				_exit(1);
			if (close(pipefds[1]))
				_exit(1);
			if (close(1))
				_exit(1);
			if (dup2(outfd, 1) < 0)
				_exit(1);
			if (close(outfd))
				_exit(1);
		}
		fpart = 1;
		if (execve(program, argv, environ))
			_exit(1);
	}
	if (errno)
		goto err;
	fpart = 0;
	if (*mode == 'r') {
		f = fdopen(pipefds[0], "r");
		if (close(pipefds[1])) {
			if (fclose(f))
				error(program);
			goto err;
		}
	} else if (*mode == 'w') {
		f = fdopen(pipefds[1], "w");
		if (close(outfd)) {
			if (fclose(f))
				error(program);
			goto err;
		}
	}
	if (!f)
		goto err;
	return f;
err:
	if (fpart)
		error(program);
	else
		error(filename);
	if (outfd >= 0 && delete(filename, 0))
		error(filename);
	if (close(pipefds[0]))
		error(filename);
	if (close(pipefds[1]))
		error(filename);
	return NULL;
}

FILE *gz_stream_open(char *filename, char *mode) {
	return compressed_stream_open(filename, mode, "/bin/gzip", "gzip");
}

FILE *bz2_stream_open(char *filename, char *mode) {
	return compressed_stream_open(filename, mode, "/bin/bzip2", "bzip2");
}

FILE *xz_stream_open(char *filename, char *mode) {
	return compressed_stream_open(filename, mode, "/bin/xz", "xz");
}

size_t xread(FILE *f, char *filename, char *buf, size_t size) {
	size_t result;
	size_t i;
	result = fread(buf, 1, size, f);
	if (ferror(f)) {
		error(filename);
		exit(1);
	}
	return result;
}

void xwrite(FILE *f, char *filename, char *buf, size_t size) {
	if (size && fwrite(buf, 1, size, f) < size) {
		error(filename);
		exit(1);
	}
}

void xseek(FILE *f, char *filename, long offset) {
	char buf[1024];
	long i;
	for (i = 0; i < offset; i += 1024) {
		if (i + 1024 < offset) {
			if (fread(buf, 1, 1024, f) < 1024)
				goto err;
		} else if (fread(buf, 1, offset - i, f) + i < offset) {
				goto err;
		}
	}
	return;
err:
	if (ferror(f))
		error(filename);
	else
		fprintf(stderr, "%s: %s: short read\n", progname,
			filename);
	exit(1);
}

size_t copy_all(FILE *in, char *infn, FILE *out, char *outfn) {
	char buf[1024];
	size_t size, total;
	total = 0;
	for (size = fread(buf, 1, 1024, in); size;
			size = fread(buf, 1, 1024, in)) {
		if (fwrite(buf, 1, size, out) < size) {
			error(outfn);
			return total;
		}
		total += size;
	}
	return total;
}

size_t x_copy_all(FILE *in, char *infn, FILE *out, char *outfn) {
	size_t result;
	result = copy_all(in, infn, out, outfn);
	if (ferror(in) || ferror(out))
		exit(1);
	return result;
}

void fill_bytes(FILE *f, char *filename, size_t size) {
	char buf[1024];
	size_t i;
	for (i = 0; i < 1024; i++)
		buf[i] = 0;
	for (i = 0; i < size; i += 1024) {
		if (i + 1024 < size)
			xwrite(f, filename, buf, 1024);
		else
			xwrite(f, filename, buf, size - i);
	}
}

void target_emsg(char *target, char *filename, char *emsg) {
	if (!target)
		fprintf(stderr, "%s: %s: %s\n", progname, filename, emsg);
	else if (target[strlen(target) - 1] == '/')
		fprintf(stderr, "%s: %s%s: %s\n", progname, target, filename, emsg);
	else
		fprintf(stderr, "%s: %s/%s: %s\n", progname, target, filename, emsg);
}

void target_error(char *target, char *filename) {
	target_emsg(target, filename, strerror(errno));
}

void target_die(char *target, char *filename) {
	target_error(target, filename);
	exit(1);
}

void version(char *cmd) {
	fprintf(stderr, "Tuesday Software Package Manager Version 5: %s\n", cmd);
}
