#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>

#include "lib.h"
#include "fs.h"
#include "archive.h"
#include "package.h"

int output_escaped_string(FILE *f, char *s) {
	for (; *s; s++) {
		switch (*s) {
		case '\\':
			if (fputc('\\', f) < 0 || fputc('\\', f) <0)
				return -1;
			break;
		case '\n':
			if (fputc('\\', f) < 0 || fputc('n', f) < 0)
				return -1;
			break;
		default:
			if (fputc(*s, f) < 0)
				return -1;
			break;
		}
	}
	return 0;
}

int write_kvp(FILE *f, char *key, char *value) {
	if (output_escaped_string(f, key))
		return -1;
	if (fputc(':', f) < 0 || fputc(' ', f) < 0)
		return -1;
	if (output_escaped_string(f, value))
		return -1;
	if (fputc('\n', f) < 0)
		return -1;
	return 0;
}

int read_kvp(FILE *f, char *filename, char **key, char **value) {
	size_t key_size, value_size;
	int mode, ch, pos;
	key_size = 16;
	*key = xmalloc(key_size, filename);
	value_size = 16;
	*value = xmalloc(value_size, filename);
	mode = 0;
	pos = 0;
	for (ch = fgetc(f); ch >= 0 && ch != 10; ch = fgetc(f)) {
		switch (mode) {
		case 0:
			switch (ch) {
			case ':':
				(*key)[pos] = 0;
				mode = 1;
				break;
			case '\\':
				mode = 3;
				break;
			default:
				(*key)[pos] = ch;
				pos++;
			}
			if (pos == key_size) {
				key_size *= 2;
				*key = xrealloc(*key, key_size, filename);
			}
			break;
		case 1:
			switch (ch) {
			case ' ': case '\t':
				break;
			case '\\':
				pos = 0;
				mode = 4;
				break;
			default:
				(*value)[0] = ch;
				pos = 1;
				mode = 2;
			}
			break;
		case 2:
			if (ch == '\\') {
				mode = 4;
			} else {
				(*value)[pos] = ch;
				pos++;
				if (pos == value_size) {
					value_size *= 2;
					*value = xrealloc(*value, value_size, filename);
				}
			}
			break;
		case 3:
			if (ch == 'n')
				(*key)[pos] = '\n';
			else
				(*key)[pos] = ch;
			pos++;
			if (pos == key_size) {
				key_size *= 2;
				*key = xrealloc(*key, key_size, filename);
			}
			mode = 0;
			break;
		case 4:
			if (ch == 'n')
				(*value)[pos] = '\n';
			else
				(*value)[pos] = ch;
			pos++;
			if (pos == value_size) {
				value_size *= 2;
				*value = xrealloc(*value, value_size, filename);
			}
			mode = 2;
			break;
		}
	}
	if (mode == 0 || mode == 3)
		(*key)[pos] = 0;
	else if (mode == 2 || mode == 4)
		(*value)[pos] = 0;
	return mode == 2 || mode == 4;
}

int rollback(struct file_list *fl) {
	for (; fl; fl = fl->next)
		if (strcmp(fl->filename, "."))
			delete(fl->filename, 0);
}

int rollback_for_overwrite(FILE *f, char *tgtdir, struct file_list *oldfl,
		struct file_list *newfl) {
	struct file_list *fl, *ofl;
	for (fl = newfl; fl; fl = fl->next) {
		if (strcmp(fl->filename, ".")) {
			for (ofl = oldfl; ofl; ofl = ofl->next) {
				if (paths_equal(fl->filename, ofl->filename) == 1) {
					if (ofl->backup && move(ofl->backup, fl->filename))
						target_error(tgtdir, fl->filename);
					else
						goto rbkdone;
				}
			}
			if (strcmp(fl->filename, "."))
				delete(fl->filename, 0);
rbkdone:
			continue;
		}
	}
	for (fl = oldfl; fl; fl = fl->next) {
		if (fl->dir) {
			if (write_kvp(f, "Directory", fl->filename))
				return -1;
		} else {
			if (write_kvp(f, "File", fl->filename))
				return -1;
		}
	}
	return 0;
}

int cleanup_overwrite(char *tgtdir, struct file_list *oldfl) {
	struct file_list *fl;
	char *name;
	int result;
	result = 0;
	for (fl = oldfl; fl; fl = fl->next) {
		if (fl->backup)
			name = fl->backup;
		else
			name = fl->filename;
		if (strcmp(name, ".") && delete(name, 0) && errno != ENOTEMPTY &&
				errno != EBUSY) {
			target_error(tgtdir, name);
			result = -1;
		}
	}
	return result;
}

FILE *open_package_datafile(char *name, char *target, char **filename) {
	FILE *f;
	*filename = malloc(10 + strlen(name));
	if (!*filename) {
		fprintf(stderr, "%s: out of memory\n", progname);
		exit(1);
	}
	sprintf(*filename, "etc/tspm/%s", name);
	f = fopen(*filename, "r");
	if (!f) {
		if (errno == ENOENT)
			return NULL;
		else
			target_die(target, *filename);
	}
	return f;
}

void cleanup_package_datafile(FILE *f, char *target, char *filename) {
	if (fclose(f))
		target_die(target, filename);
	free(filename);
}

char *read_dep_var(char *filename, char *depdata, size_t *i) {
	char *result;
	if (depdata[*i]) {
		result = strdup(depdata + *i);
		if (!result) {
			fprintf(stderr, "%s: %s: out of memory\n", progname, filename);
			exit(1);
		}
		if (strchr(result, 10)) {
			fprintf(stderr, "%s: %s: invalid TSP file: "
				"dependency information contains newline\n", progname,
				filename);
			exit(1);
		}
		(*i) += strlen(result) + 1;
		return result;
	} else {
		(*i)++;
		return NULL;
	}
}

char *dup_package_header_var(char *src, char *filename, size_t size) {
	char *result;
	size_t len;
	if (!*src)
		return NULL;
	for (len = 0; src[len] && len < size; len++);
	result = xmalloc(len + 1, filename);
	memcpy(result, src, len);
	result[len] = 0;
	if (strchr(result, 10)) {
		fprintf(stderr,
			"%s: invalid TSP file: package version contains newline\n",
			progname);
		exit(1);
	}
	return result;
}

void read_package_header(FILE *in, char *infn, char **name, char **vers,
		struct dep **deps) {
	struct dep *new, *cd;
	char buf[512];
	char *depdata;
	size_t size, tarsize, i;
	if (xread(in, infn, buf, 512) < 512) {
		fprintf(stderr, "%s: %s: invalid TSP file: short read\n", progname,
			infn);
		exit(1);
	}
	if (buf[156] != '#') {
		fprintf(stderr, "%s: %s: invalid TSP file: bad header type\n",
			progname, infn);
		exit(1);
	}
	if (!checksum_is_valid(buf, infn))
		exit(1);
	size = strtol(buf + 124, NULL, 8);
	tarsize = ((size + 511) & ~511);
	depdata = xmalloc(tarsize, infn);
	cd = NULL;
	if (xread(in, infn, depdata, tarsize) < tarsize) {
		fprintf(stderr, "%s: %s: invalid TSP file: short read\n", progname,
			infn);
		exit(1);
	}
	*deps = NULL;
	for (i = 0; i < size;) {
		new = xmalloc(sizeof(struct dep), infn);
		new->name = read_dep_var(infn, depdata, &i);
		if (!new->name) {
			fprintf(stderr, "%s: invalid TSP file: dependency has no name\n",
				progname);
			exit(1);
		}
		new->min_version = read_dep_var(infn, depdata, &i);
		new->max_version = read_dep_var(infn, depdata, &i);
		new->next = NULL;
		if (!cd)
			*deps = new;
		else
			cd->next = new;
		cd = new;
	}
	*name = dup_package_header_var(buf, infn, 100);
	if (!*name) {
		fprintf(stderr, "%s: invalid TSP file: package has no name\n",
			progname);
		exit(1);
	}
	*vers = dup_package_header_var(buf + 100, infn, 24);
	if (!*vers) {
		fprintf(stderr, "%s: invalid TSP file: package has no version\n",
			progname);
		exit(1);
	}
}
