#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <fcntl.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>

#include "lib.h"
#include "archive.h"

void usage() {
	fprintf(stderr,
		"\n%s -x [OPTIONS] [FILENAME]\n"
		"\tExtract the archive on standard input or in FILENAME\n"
		"%s -t [OPTIONS] [FILENAME]\n"
		"\tList the files in the archive on standard input or in FILENAME\n"
		"%s -c [OPTIONS] [FILENAME] [INPUT ...]\n"
		"\tCreate a new archive on standard output or in FILENAME containing\n"
		"\tthe INPUT files\n"
		"%s -e\n"
		"%s --version\n"
		"\tDisplay the version of this command\n"
		"%s -?\n"
		"%s --help\n"
		"\tDisplay this usage message\n"
		"If all options are - options and not -- options, and they are\n"
		"consolidated into a single argument, the - may be omitted.\n"
		"\t-x\tuse GZIP compression\n"
		"\t-j\tuse BZIP2 compression\n"
		"\t-J\tuse XZ compression\n"
		"\t-p\tpreserve permission in excess of umask\n"
		"\t-k\tdo not overwrite old files\n\n",
		progname, progname, progname, progname, progname, progname, progname);
}

#define NONE 0
#define CREATE 1
#define LIST 2
#define EXTRACT 3
#define VERSION 4
#define USAGE 5

#define GZIP 16
#define BZIP2 17
#define XZ 18

int identify_compression(char *filename) {
	FILE *f;
	int result;
	if (filename && strcmp(filename, "-")) {
		f = fopen(filename, "r");
		if (!f)
			die(filename);
	} else {
		f = stdin;
	}
	result = NONE;
	if (fseek(f, 0L, SEEK_SET)) {
		errno = 0;
		return result;
	}
	switch (fgetc(f)) {
	case 31:
		if (fgetc(f) == 139)
			result = GZIP;
		break;
	case 'B':
		if (fgetc(f) == 'Z' && fgetc(f) == 'h')
			result = BZIP2;
		break;
	case 0xfd:
		if (fgetc(f) == '7' && fgetc(f) == 'z' && fgetc(f) == 'X' &&
				fgetc(f) == 'Z' && fgetc(f) == 0)
			result = XZ;
	}
	if (f != stdin) {
		if (fclose(f))
			error(filename);
	} else {
		rewind(f);
	}
	return result;
}

int main(int argc, char **argv) {
	struct passwd *pw;
	struct group *gr;
	FILE *f;
	struct path_info *pi, *cpi;
	uid_t uid;
	gid_t gid;
	char *s, *filename, *uname, *gname, *ffn;
	int i, action, compression, extract_options, needs_archive, long_option, r;
	int result;
	progname = argv[0];
	if (argc < 2) {
		fprintf(stderr, "%s: no arguments ('%s ?' for usage)\n", progname,
			progname);
		return 1;
	}
	result = 0;
	action = NONE;
	compression = NONE;
	filename = NULL;
	extract_options = 0;
	needs_archive = 0;
	uname = NULL;
	gname = NULL;
	uid = 65534;
	gid = 65534;
	pi = NULL;
	for (i = 1; i < argc && (i == 1 || (argv[i][0] == '-' && argv[i][1]));
			i++) {
		if (*(argv[i]) == '-') {
			if (argv[i][1] == '-') {
				s = argv[i] + 2;
				long_option = 1;
			} else {
				s = argv[i] + 1;
				long_option = 0;
			}
		} else {
			s = argv[i];
			long_option = 0;
		}
		if (long_option) {
			if (!strcmp(s, "version")) {
				action = VERSION;
			} else if (!strcmp(s, "help")) {
				action = USAGE;
			} else if (!strncmp(s, "owner=", 6)) {
				uname = s + 6;
				pw = getpwnam(uname);
				if (pw)
					uid = pw->pw_uid;
				else
					fprintf(stderr, "%s: warning: invalid %s\n", progname,
						uname);
			} else if (!strncmp(s, "group=", 6)) {
				gname = s + 6;
				gr = getgrnam(gname);
				if (gr)
					gid = gr->gr_gid;
				else
					fprintf(stderr, "%s: warning: invalid %s\n", progname,
						gname);
			} else {
				fprintf(stderr, "%s: invalid --%s ('%s ?' for usage)\n",
					progname, s, progname);
				return 1;
			}
		} else {
			for (; *s; s++) {
			switch (*s) {
				case 'c':
					action = CREATE;
					break;
				case 'e':
					action = VERSION;
					break;
				case 'f':
					needs_archive = 1;
					break;
				case 'j':
					compression = BZIP2;
					break;
				case 'J':
					compression = XZ;
					break;
				case 'k':
					extract_options |= EXTRACT_KEEP;
					break;
				case 'p':
					extract_options |= EXTRACT_PERMISSION;
					break;
				case 't':
					action = LIST;
					break;
				case 'x':
					action = EXTRACT;
					break;
				case 'z':
					compression = GZIP;
					break;
				case '?':
					action = USAGE;
					break;
				default:
					fprintf(stderr, "%s: invalid -%c ('%s ?' for usage)\n",
						progname, *s, progname);
					return 1;
				}
			}
		}
	}
	if (needs_archive) {
		if (argc <= i) {
			fprintf(stderr, "%s: no filename ('%s ?' for usage)\n", progname,
				progname);
			return 1;
		}
		filename = argv[i];
		i++;
	}
	if (!filename)
		filename = "-";
	switch (action) {
	case NONE:
		fprintf(stderr, "%s: no c, t, or x ('%s ?' for usage)\n", progname,
			progname);
		return 1;
	case CREATE:
		ffn = filename;
		switch (compression) {
		case NONE:
			f = filename_adjust_open(&ffn, "w");
			break;
		case GZIP:
			f = gz_stream_open(ffn, "w");
			break;
		case BZIP2:
			f = bz2_stream_open(ffn, "w");
			break;
		case XZ:
			f = xz_stream_open(ffn, "w");
			break;
		}
		if (!f)
			return 1;
		for (; i < argc; i++)
			if (prepare_path(&pi, NULL, argv[i], 0))
				result = 1;
		for (cpi = pi; cpi; cpi = cpi->next)
			if (output_path(f, ffn, cpi, pi, NULL, uid, uname, gid, gname))
				result = 1;
		output_footer(f, ffn);
		if (fclose(f))
			error(ffn);
		break;
	case LIST: case EXTRACT:
		if (compression == NONE)
			compression = identify_compression(filename);
		ffn = filename;
		switch (compression) {
		case NONE:
			f = filename_adjust_open(&ffn, "r");
			break;
		case GZIP:
			f = gz_stream_open(ffn, "r");
			break;
		case BZIP2:
			f = bz2_stream_open(ffn, "r");
			break;
		case XZ:
			f = xz_stream_open(ffn, "r");
			break;
		}
		if (!f)
			return 1;
		if (action == LIST) {
			for (r = list_file(f, ffn, 0); r; r = list_file(f, ffn, 0))
				if (r < 0)
					result = 1;
		} else if (action == EXTRACT) {
			for (r = extract_file(NULL, NULL, NULL, extract_options, 0, 0, 0,
						NULL, f, ffn, 0); r;
					r = extract_file(NULL, NULL, NULL, extract_options, 0, 0,
						0, NULL, f, ffn, 0))
				if (r < 0)
					result = 1;
		}
		if (fclose(f))
			result = 1;
		break;
	case VERSION:
		version("tar");
		break;
	case USAGE:
		version("tar");
		usage();
		break;
	}
	return result;
}
